from json import dumps, loads
from urllib.request import Request, urlopen
from urllib.error import HTTPError, URLError
from urllib.parse import urlencode

from telecast import exceptions, codes


def api_call(url, **kwargs):
    request = Request(url, dumps(kwargs).encode(), headers={
        'Content-Type': 'application/json'
    })
    try:
        response = urlopen(request)
        data = response.read()
        status = response.status
    except HTTPError as error:
        data = error.read()
        status = error.code
    except URLError as error:
        raise exceptions.RPCProtocolError(str(error))

    if status != 200:
        raise exceptions.RPCProtocolError('HTTP ' + str(status) + ': ' + str(data))

    try:
        data = loads(data)
        result = data['result']
        code = data['code']
    except Exception as error:
        raise exceptions.RPCProtocolError(str(error))
    else:
        if code == codes.RPC_CODE_PROTOCOL_ERROR:
            # Not used at the moment.
            raise exceptions.RPCProtocolError(str(result))
        if code == codes.RPC_CODE_NOT_ALLOWED:
            raise exceptions.RPCNotAllowedError(str(result))
        if code == codes.RPC_CODE_REMOTE_ERROR:
            raise exceptions.RPCRemoteError(str(result))
        return result



import inspect

import coreapi
import coreschema
from django.conf import settings
from rest_framework.decorators import api_view, schema
from rest_framework.response import Response
from rest_framework.schemas import AutoSchema

from telecast import codes

IGNORED_FIELDS = ['request']


def get_is_port_allowed(port):
    """
    Return TELECAST_PORTS.
    """
    if not hasattr(settings, 'TELECAST_PORTS'):
        return True
    ports = settings.TELECAST_PORTS
    if not isinstance(ports, (list, tuple)):
        ports = [ports]
    return str(port) in map(str, ports)


def method(**decorator_kwargs):
    """
    Return decorator that mimics api_view.
    """
    def decorator(func):
        """
        Decorate a method with api_view and add proper schema that matches
        function arguments and return a proxy that transforms this function
        into a callable REST endpoint.
        """
        if 'http_method_names' not in decorator_kwargs:
            decorator_kwargs['http_method_names'] = ['POST']

        @api_view(**decorator_kwargs)
        @schema(AutoSchema(manual_fields=[
            coreapi.Field(
                field,
                required=False,
                location='form',
                schema=coreschema.Anything()
            )
            for field
            in inspect.signature(func).parameters
            if field not in IGNORED_FIELDS
        ]))
        def wrapper(request, *args, **kwargs):
            """
            Proxy that wraps a function into a REST endpoint.
            """
            assert not args
            assert not kwargs
            port = request.META.get('HTTP_X_FORWARDED_PORT')
            if port and not get_is_port_allowed(port):
                return Response(dict(
                    code=codes.RPC_CODE_NOT_ALLOWED,
                    result='You are not allowed to call this TELECAST method.'
                ))

            if hasattr(request, 'data'):
                arguments = request.data
            else:
                # e. g. GET request
                arguments = {}

            try:
                result = func(request, **arguments)
            except Exception as error:
                return Response(dict(
                    code=codes.RPC_CODE_REMOTE_ERROR,
                    result=str(error)
                ))

            return Response(dict(
                code=codes.RPC_CODE_OK,
                result=result
            ))
        return wrapper
    return decorator



from django.conf import settings

from telecast.client import api_call


def call(path, **kwargs):
    assert hasattr(settings, 'TELECAST_URL'), 'TELECAST_URL is required to perform RPC calls.'
    url = settings.TELECAST_URL.rstrip('/') + '/' + path.lstrip('/')
    return api_call(url, **kwargs)



class RPCError(Exception):
    """
    Generic RPC exception.
    Thrown when an RPC request was not completed for some reason.
    """
    pass


class RPCRemoteError(RPCError):
    """
    Should be raised manually by developer from an RPC method
    if a predictable exception occurs that should be propagated to a caller.
    This exception will be serialized, sent as response and re-raised
    on caller side with the same message.
    """
    pass


class RPCNotAllowedError(RPCError):
    """
    Raised on server end and reraised on client end if caller
    is not allowed to call a specific method.
    """


class RPCProtocolError(RPCError):
    """
    Raised on client end if an unknown response was received.
    Indicates that the called URL does not represent
    a valid RPC endpoint or that the callee network was down.
    """
    pass



#!/usr/bin/env python3.6

from setuptools import setup
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='telecast',
    version='0.3',
    # use_scm_version=True,
    # setup_requires=['setuptools_scm'],
    description='Portable REST/JSON RPC implementation.',
    long_description=long_description,
    url='https://github.com/and3rson/telecast',
    author='Andrew Dunai',
    author_email='a@dun.ai',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='python json rest rpc',
    packages=[
        'telecast',
        'telecast.contrib',
        'telecast.contrib.django',
        'telecast.contrib.rest_framework'
    ],
)
